import datetime
from io import BytesIO
from unittest import mock

from django.conf import settings
from django.conf.locale import LANG_INFO
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import F
from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from django.urls import reverse
from django.utils import timezone, translation
from freezegun import freeze_time
from openpyxl import load_workbook

from wagtail.admin.views.mixins import ExcelDateFormatter
from wagtail.admin.views.reports import page_types_usage
from wagtail.admin.views.reports.audit_logging import LogEntriesView
from wagtail.models import (
    GroupPagePermission,
    Locale,
    ModelLogEntry,
    Page,
    PageLogEntry,
    Site,
)
from wagtail.test.testapp.models import (
    Advert,
    EventPage,
    EventPageSpeaker,
    SimplePage,
)
from wagtail.test.utils import WagtailTestUtils
from wagtail.test.utils.template_tests import AdminTemplateTestUtils


class BaseReportViewTestCase(AdminTemplateTestUtils, WagtailTestUtils, TestCase):
    url_name = None
    header_buttons_parent_selector = "#w-slim-header-buttons"
    drilldown_selector = ".w-drilldown"
    extra_params = ""
    results_only = False

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.url = reverse(cls.url_name)
        if cls.results_only:
            cls.header_buttons_parent_selector = (
                '[data-controller="w-teleport"]'
                '[data-w-teleport-target-value="#w-slim-header-buttons"]'
            )
            cls.drilldown_selector = (
                '[data-controller="w-teleport"]'
                '[data-w-teleport-target-value="#filters-drilldown"]'
            )
            cls.extra_params = "&_w_filter_fragment=true"

    def setUp(self):
        self.user = self.login()

    def get(self, params={}, **kwargs):
        if self.results_only:
            params["_w_filter_fragment"] = "true"
        return self.client.get(self.url, params, **kwargs)

    def assertActiveFilter(self, soup, name, value):
        # Should render the export buttons inside the header "more" dropdown
        # with the filtered URL. When used in a results-only view, these are
        # teleported to the correct element in the skeleton.
        links_parent = soup.select_one(self.header_buttons_parent_selector)
        self.assertIsNotNone(links_parent)
        links = links_parent.select(".w-dropdown a")
        unfiltered_url = reverse(self.url_name)
        filtered_url = f"{unfiltered_url}?{name}={value}{self.extra_params}"
        self.assertEqual(len(links), 2)
        self.assertEqual(
            [link.get("href") for link in links],
            [f"{filtered_url}&export=xlsx", f"{filtered_url}&export=csv"],
        )

        # Should render the active filter pill
        active_filter = soup.select_one(".w-active-filters .w-pill__content")
        clear_button = soup.select_one(".w-active-filters .w-pill__remove")
        self.assertIsNotNone(active_filter)
        self.assertIsNotNone(clear_button)
        self.assertNotIn(name, clear_button.attrs.get("data-w-swap-src-value"))
        self.assertEqual(clear_button.attrs.get("data-w-swap-reflect-value"), "true")

    def assertActiveFilterNotRendered(self, soup):
        self.assertIsNone(soup.select_one(".w-active-filters"))

    def assertBreadcrumbs(self, breadcrumbs, html):
        if self.results_only:
            self.assertBreadcrumbsNotRendered(html)
        else:
            self.assertBreadcrumbsItemsRendered(breadcrumbs, html)

    def assertPageTitle(self, soup, title):
        page_title = soup.select_one("title")
        if self.results_only:
            self.assertIsNone(page_title)
        else:
            self.assertIsNotNone(page_title)
            self.assertEqual(page_title.text.strip(), title)


class TestLockedPagesView(BaseReportViewTestCase):
    url_name = "wagtailadmin_reports:locked_pages"

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(
            response,
            "wagtailadmin/reports/base_page_report.html",
        )
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/locked_pages_results.html",
        )
        self.assertBreadcrumbs(
            [{"url": "", "label": "Locked pages"}],
            response.content,
        )

        # Initially there should be no locked pages
        self.assertContains(response, "No locked pages found.")

        # Should render the filter inside the drilldown
        soup = self.get_soup(response.content)
        locked_by_options = soup.select(
            f"{self.drilldown_selector} select[name='locked_by'] option"
        )
        # No user locked anything, so there should be no option for the filter
        self.assertEqual(len(locked_by_options), 1)
        self.assertEqual(locked_by_options[0].text, "---------")
        self.assertEqual(locked_by_options[0].get("value"), "")
        self.assertActiveFilterNotRendered(soup)
        self.assertPageTitle(soup, "Locked pages - Wagtail")

        parent_page = Page.objects.first()
        parent_page.add_child(
            instance=Page(
                title="First locked page",
                locked=True,
                locked_by=self.user,
                locked_at=timezone.now(),
            )
        )
        parent_page.add_child(
            instance=Page(
                title="Second locked page",
                locked=True,
                locked_by=self.user,
                locked_at=timezone.now(),
            )
        )

        # Now the listing should contain our locked page
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(
            response,
            "wagtailadmin/reports/base_page_report.html",
        )
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/locked_pages_results.html",
        )
        self.assertBreadcrumbs(
            [{"url": "", "label": "Locked pages"}],
            response.content,
        )
        self.assertNotContains(response, "No locked pages found.")
        self.assertContains(response, "First locked page")
        self.assertContains(response, "Second locked page")

        # Should render the filter inside the drilldown
        soup = self.get_soup(response.content)
        locked_by_options = soup.select(
            f"{self.drilldown_selector} select[name='locked_by'] option"
        )
        # The options should only display users who have locked pages
        self.assertEqual(len(locked_by_options), 2)
        self.assertEqual(locked_by_options[0].text, "---------")
        self.assertIsNone(locked_by_options[0].value)
        self.assertEqual(locked_by_options[1].text, str(self.user))
        self.assertEqual(locked_by_options[1].get("value"), str(self.user.pk))
        self.assertActiveFilterNotRendered(soup)

        # Locked by current user shown in indicator
        self.assertNotContains(response, "indicator--is-dimmed")
        self.assertContains(
            response, 'title="This page is locked, by you, to further editing"'
        )

    def test_get_with_minimal_permissions(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        GroupPagePermission.objects.create(
            group=group,
            page=Page.objects.first(),
            permission_type="unlock",
        )

        response = self.get()

        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(
            response,
            "wagtailadmin/reports/base_page_report.html",
        )
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/locked_pages_results.html",
        )
        self.assertContains(response, "No locked pages found.")

    def test_get_with_no_permissions(self):
        self.user.is_superuser = False
        self.user.save()
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )

        response = self.get()

        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_csv_export(self):
        self.page = Page.objects.first()
        self.page.locked = True
        self.page.locked_by = self.user
        if settings.USE_TZ:
            # 12:00 UTC
            self.page.locked_at = "2013-02-01T12:00:00.000Z"
            self.page.latest_revision_created_at = "2013-01-01T12:00:00.000Z"
        else:
            # 12:00 in no specific timezone
            self.page.locked_at = "2013-02-01T12:00:00"
            self.page.latest_revision_created_at = "2013-01-01T12:00:00"
        self.page.save()

        response = self.get(params={"export": "csv"})

        # Check response
        self.assertEqual(response.status_code, 200)
        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(
            data_lines[0], "Title,Updated,Status,Type,Locked at,Locked by\r"
        )
        if settings.USE_TZ:
            self.assertEqual(
                data_lines[1],
                "Root,2013-01-01 12:00:00+00:00,live,Page,2013-02-01 12:00:00+00:00,test@email.com\r",
            )
        else:
            self.assertEqual(
                data_lines[1],
                "Root,2013-01-01 12:00:00,live,Page,2013-02-01 12:00:00,test@email.com\r",
            )

    def test_xlsx_export(self):
        self.page = Page.objects.first()
        self.page.locked = True
        self.page.locked_by = self.user
        if settings.USE_TZ:
            # 12:00 UTC
            self.page.locked_at = "2013-02-01T12:00:00.000Z"
            self.page.latest_revision_created_at = "2013-01-01T12:00:00.000Z"
        else:
            # 12:00 in no specific timezone
            self.page.locked_at = "2013-02-01T12:00:00"
            self.page.latest_revision_created_at = "2013-01-01T12:00:00"
        self.page.save()

        response = self.get(params={"export": "xlsx"})

        # Check response - the locked page info should be in it
        self.assertEqual(response.status_code, 200)
        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))["Sheet1"]
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]
        self.assertEqual(
            cell_array[0],
            ["Title", "Updated", "Status", "Type", "Locked at", "Locked by"],
        )
        self.assertEqual(
            cell_array[1],
            [
                "Root",
                datetime.datetime(2013, 1, 1, 12, 0),
                "live",
                "Page",
                datetime.datetime(2013, 2, 1, 12, 0),
                "test@email.com",
            ],
        )
        self.assertEqual(len(cell_array), 2)

        self.assertEqual(worksheet["B2"].number_format, ExcelDateFormatter().get())
        self.assertEqual(worksheet["E2"].number_format, ExcelDateFormatter().get())


class TestFilteredLockedPagesView(BaseReportViewTestCase):
    fixtures = ["test.json"]
    url_name = "wagtailadmin_reports:locked_pages"

    def setUp(self):
        self.user = self.login()
        self.unpublished_page = Page.objects.get(
            url_path="/home/events/tentative-unpublished-event/"
        )
        self.unpublished_page.locked = True
        self.unpublished_page.locked_by = self.user
        self.unpublished_page.locked_at = timezone.now()
        self.unpublished_page.save()

        self.christmas_page = Page.objects.get(url_path="/home/events/christmas/")
        self.christmas_page.locked = True
        self.christmas_page.locked_by = self.user
        self.christmas_page.locked_at = timezone.now()
        self.christmas_page.save()

    def test_filter_by_live(self):
        response = self.get(params={"live": "true"})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, "Tentative Unpublished Event")
        self.assertContains(response, "My locked page")
        self.assertContains(response, "Christmas")

        response = self.get(params={"live": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Tentative Unpublished Event")
        self.assertNotContains(response, "My locked page")
        self.assertNotContains(response, "Christmas")

        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "live", "false")

    def test_filter_by_user(self):
        response = self.get(params={"locked_by": self.user.pk})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Tentative Unpublished Event")
        self.assertContains(response, "Christmas")
        self.assertNotContains(response, "My locked page")


class TestFilteredLockedPagesResultsView(TestFilteredLockedPagesView):
    url_name = "wagtailadmin_reports:locked_pages_results"
    results_only = True


class TestFilteredLogEntriesView(BaseReportViewTestCase):
    fixtures = ["test.json"]
    url_name = "wagtailadmin_reports:site_history"

    def setUp(self):
        self.user = self.login()
        self.home_page = Page.objects.get(url_path="/home/")
        self.custom_model = Advert.objects.get(pk=1)

        self.editor = self.create_user(
            username="the_editor", email="the_editor@example.com", password="password"
        )
        editors = Group.objects.get(name="Editors")
        editors.permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        GroupPagePermission.objects.create(
            group=editors, page=self.home_page, permission_type="change"
        )
        editors.user_set.add(self.editor)

        # timezone matches TIME_ZONE = "Asia/Tokyo" in tests/settings.py
        with freeze_time("2024-05-06 12:00:00+09:00"):
            self.today = timezone.now()

        self.create_log = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.create",
            timestamp=self.today - timezone.timedelta(days=4),
            user=self.user,
        )
        self.edit_log_1 = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.edit",
            timestamp=self.today - timezone.timedelta(days=3),
        )
        self.edit_log_2 = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.edit",
            timestamp=self.today - timezone.timedelta(days=2),
            user=self.editor,
        )
        self.edit_log_3 = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.edit",
            timestamp=self.today - timezone.timedelta(days=1),
            title="The FINAL cut",
        )

        self.create_comment_log = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.comments.create",
            data={
                "comment": {
                    "contentpath": "title",
                    "text": "Foo",
                }
            },
        )
        self.edit_comment_log = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.comments.edit",
            data={
                "comment": {
                    "contentpath": "title",
                    "text": "Edited",
                }
            },
        )
        self.create_reply_log = PageLogEntry.objects.log_action(
            self.home_page,
            "wagtail.comments.create_reply",
            data={
                "comment": {
                    "contentpath": "title",
                    "text": "Foo",
                }
            },
        )

        self.create_custom_log = ModelLogEntry.objects.log_action(
            self.custom_model,
            "wagtail.create",
            timestamp=self.today - timezone.timedelta(days=3),
        )

        self.edit_custom_log = ModelLogEntry.objects.log_action(
            self.custom_model,
            "wagtail.edit",
            timestamp=self.today - timezone.timedelta(days=2),
            title="the final CUT",
        )

    def assert_log_entries(self, response, expected):
        actual = set(response.context["object_list"])
        self.assertSetEqual(actual, set(expected))

    def assert_filter_actions(self, response, expected):
        soup = self.get_soup(response.content)
        actual = {
            choice.get("value")
            for choice in soup.select(
                f"{self.drilldown_selector} input[name='action'][type='checkbox']"
            )
        }
        self.assertSetEqual(actual, set(expected))

    def test_unfiltered(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertBreadcrumbs(
            [{"url": "", "label": "Site history"}],
            response.content,
        )
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )

        self.assert_filter_actions(
            response,
            [
                "wagtail.create",
                "wagtail.edit",
                "wagtail.comments.create",
                "wagtail.comments.edit",
                "wagtail.comments.create_reply",
            ],
        )

        soup = self.get_soup(response.content)
        self.assertActiveFilterNotRendered(soup)
        self.assertPageTitle(soup, "Site history - Wagtail")

        # The editor should not see the Advert's log entries.
        self.login(user=self.editor)
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
            ],
        )

        self.assert_filter_actions(
            response,
            [
                "wagtail.create",
                "wagtail.edit",
                "wagtail.comments.create",
                "wagtail.comments.edit",
                "wagtail.comments.create_reply",
            ],
        )

        soup = self.get_soup(response.content)
        self.assertActiveFilterNotRendered(soup)

    def test_filter_by_action(self):
        response = self.get(params={"action": "wagtail.edit"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.edit_custom_log,
            ],
        )

        self.login(user=self.editor)
        response = self.get(params={"action": "wagtail.edit"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
            ],
        )

        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "action", "wagtail.edit")

    def test_filter_by_action_multiple(self):
        response = self.get(params={"action": ["wagtail.edit", "wagtail.create"]})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.create_custom_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.edit_custom_log,
            ],
        )

        self.login(user=self.editor)
        response = self.get(params={"action": ["wagtail.edit", "wagtail.create"]})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
            ],
        )

    def test_filter_by_timestamp(self):
        today = self.today.date()
        response = self.get(
            params={"timestamp_from": today - timezone.timedelta(days=3)}
        )
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                # Doesn't contain self.create_log which was created 4 days ago
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )

        response = self.get(params={"timestamp_to": today - timezone.timedelta(days=2)})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                # Doesn't contain self.edit_log_3 which was created 1 day ago,
                # as well as self.create_comment_log, self.edit_comment_log,
                # and self.create_reply_log which was created without an explicit
                # timestamp (and thus defaults to the current time)
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )

        response = self.get(
            params={
                "timestamp_from": today - timezone.timedelta(days=3),
                "timestamp_to": today - timezone.timedelta(days=2),
            }
        )
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                # Doesn't contain
                # self.create_log which was created 4 days ago,
                # self.edit_log_3 which was created 1 day ago,
                # as well as self.create_comment_log, self.edit_comment_log,
                # and self.create_reply_log which was created without an explicit
                # timestamp (and thus defaults to the current time)
                self.edit_log_1,
                self.edit_log_2,
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )

    def test_filter_by_user(self):
        response = self.get(params={"user": self.editor.pk})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(response, [self.edit_log_2])

        response = self.get(params={"user": [self.user.pk, self.editor.pk]})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(response, [self.create_log, self.edit_log_2])

    def test_filter_by_label(self):
        response = self.get(params={"label": "final cut"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(response, [self.edit_log_3, self.edit_custom_log])

    def test_filter_by_object_type(self):
        response = self.get(
            params={"object_type": ContentType.objects.get_for_model(Page).pk}
        )
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
            ],
        )

        response = self.get(
            params={"object_type": ContentType.objects.get_for_model(Advert).pk}
        )
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )

    def test_is_commenting_action(self):
        response = self.get(params={"is_commenting_action": "false"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
                self.create_custom_log,
                self.edit_custom_log,
            ],
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "is_commenting_action", "false")

        response = self.get(params={"is_commenting_action": "true"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
            ],
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "is_commenting_action", "true")

        self.login(user=self.editor)
        response = self.get(params={"is_commenting_action": "false"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_log,
                self.edit_log_1,
                self.edit_log_2,
                self.edit_log_3,
            ],
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "is_commenting_action", "false")

        response = self.get(params={"is_commenting_action": "true"})
        self.assertEqual(response.status_code, 200)
        self.assert_log_entries(
            response,
            [
                self.create_comment_log,
                self.edit_comment_log,
                self.create_reply_log,
            ],
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "is_commenting_action", "true")

    def test_log_entry_with_stale_content_type(self):
        stale_content_type = ContentType.objects.create(
            app_label="fake_app", model="deleted model"
        )

        ModelLogEntry.objects.create(
            object_id=123,
            content_type=stale_content_type,
            label="This instance's model was deleted, but its content type was not",
            action="wagtail.create",
            timestamp=timezone.now(),
        )

        response = self.get()
        self.assertContains(response, "Deleted model")

    def test_log_entry_with_null_content_type(self):
        ModelLogEntry.objects.create(
            object_id=123,
            content_type=None,
            label="This instance's model was deleted, and so was its content type",
            action="wagtail.create",
            timestamp=timezone.now(),
        )

        response = self.get()
        self.assertContains(response, "Unknown content type")

    def test_decorated_queryset(self):
        # Ensure that decorate_paginated_queryset is only called with the queryset for the current
        # page, instead of all objects over all pages.
        with (
            mock.patch.object(
                LogEntriesView,
                "decorate_paginated_queryset",
                side_effect=LogEntriesView.decorate_paginated_queryset,
                autospec=True,
            ) as decorate_paginated_queryset,
            mock.patch.object(LogEntriesView, "paginate_by", return_value=1),
        ):
            response = self.get()
            decorate_paginated_queryset.assert_called_once()
            queryset = decorate_paginated_queryset.call_args.args[1]
            self.assertEqual(queryset.count(), 1)

        self.assertEqual(response.status_code, 200)


class TestFilteredLogEntriesResultsView(TestFilteredLogEntriesView):
    url_name = "wagtailadmin_reports:site_history_results"
    results_only = True


@override_settings(
    USE_L10N=True,
)
class TestExcelDateFormatter(TestCase):
    def test_all_locales(self):
        formatter = ExcelDateFormatter()

        for lang in LANG_INFO.keys():
            with self.subTest(lang), translation.override(lang):
                self.assertNotEqual(formatter.get(), "")

    def test_format(self):
        formatter = ExcelDateFormatter()

        with self.subTest(format="r"):
            # Format code for RFC 5322 formatted date, e.g. 'Thu, 21 Dec 2000 16:01:07'
            self.assertEqual(formatter.format("r"), "ddd, d mmm yyyy hh:mm:ss")

        with self.subTest(format="m/d/Y g:i A"):
            # Format code for e.g. '12/21/2000 4:01 PM'
            self.assertEqual(formatter.format("m/d/Y g:i A"), "mm/dd/yyyy h:mm AM/PM")


class TestAgingPagesView(BaseReportViewTestCase):
    url_name = "wagtailadmin_reports:aging_pages"

    def setUp(self):
        self.user = self.login()
        self.root = Page.objects.first()
        self.home = Page.objects.get(slug="home")

    def publish_home_page(self):
        self.home.save_revision().publish(user=self.user)

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateNotUsed(
            response,
            "wagtailadmin/reports/base_page_report.html",
        )
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/aging_pages_results.html",
        )
        self.assertBreadcrumbs(
            [{"url": "", "label": "Aging pages"}],
            response.content,
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilterNotRendered(soup)
        self.assertPageTitle(soup, "Aging pages - Wagtail")

    def test_displays_only_published_pages(self):
        response = self.get()
        self.assertContains(response, "No pages found.")

        self.publish_home_page()
        response = self.get()

        # Home Page should be listed
        self.assertContains(response, self.home.title)
        # Last published by user is set
        self.assertContains(response, self.user.get_username())

        self.assertNotContains(response, self.root.title)
        self.assertNotContains(response, "No pages found.")

    def test_permissions(self):
        # Publish home page
        self.publish_home_page()

        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.get()
        self.assertEqual(response.status_code, 302)

    def test_csv_export(self):
        self.publish_home_page()
        if settings.USE_TZ:
            self.home.last_published_at = "2013-01-01T12:00:00.000Z"
        else:
            self.home.last_published_at = "2013-01-01T12:00:00"
        self.home.save()

        response = self.get(params={"export": "csv"})
        self.assertEqual(response.status_code, 200)

        data_lines = response.getvalue().decode().split("\n")
        self.assertEqual(
            data_lines[0], "Title,Status,Last published at,Last published by,Type\r"
        )
        if settings.USE_TZ:
            self.assertEqual(
                data_lines[1],
                "Welcome to your new Wagtail site!,live + draft,2013-01-01 12:00:00+00:00,test@email.com,Page\r",
            )
        else:
            self.assertEqual(
                data_lines[1],
                "Welcome to your new Wagtail site!,live + draft,2013-01-01 12:00:00,test@email.com,Page\r",
            )

    def test_xlsx_export(self):
        self.publish_home_page()

        if settings.USE_TZ:
            self.home.last_published_at = "2013-01-01T12:00:00.000Z"
        else:
            self.home.last_published_at = "2013-01-01T12:00:00"
        self.home.save()

        response = self.get(params={"export": "xlsx"})
        self.assertEqual(response.status_code, 200)

        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))["Sheet1"]
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]

        self.assertEqual(
            cell_array[0],
            ["Title", "Status", "Last published at", "Last published by", "Type"],
        )
        self.assertEqual(
            cell_array[1],
            [
                "Welcome to your new Wagtail site!",
                "live + draft",
                datetime.datetime(2013, 1, 1, 12, 0),
                "test@email.com",
                "Page",
            ],
        )
        self.assertEqual(len(cell_array), 2)

        self.assertEqual(worksheet["C2"].number_format, ExcelDateFormatter().get())

    def test_xlsx_export_without_published_by(self):
        """
        Test that the xlsx export works when a page has no 'published_by' set.
        See https://github.com/wagtail/wagtail/issues/10821
        """

        self.home.save_revision().publish()

        if settings.USE_TZ:
            self.home.last_published_at = "2013-01-01T12:00:00.000Z"
        else:
            self.home.last_published_at = "2013-01-01T12:00:00"

        # mimic a page that does not have a 'published_by' on creation
        self.home.last_published_by = None
        self.home.save()

        response = self.get(params={"export": "xlsx"})
        self.assertEqual(response.status_code, 200)

        workbook_data = response.getvalue()
        worksheet = load_workbook(filename=BytesIO(workbook_data))["Sheet1"]
        cell_array = [[cell.value for cell in row] for row in worksheet.rows]

        self.assertEqual(
            cell_array[0],
            ["Title", "Status", "Last published at", "Last published by", "Type"],
        )
        self.assertEqual(
            cell_array[1],
            [
                "Welcome to your new Wagtail site!",
                "live + draft",
                datetime.datetime(2013, 1, 1, 12, 0),
                None,
                "Page",
            ],
        )
        self.assertEqual(len(cell_array), 2)

        self.assertEqual(worksheet["C2"].number_format, ExcelDateFormatter().get())

    def test_report_renders_when_page_publisher_deleted(self):
        temp_user = self.create_superuser(
            "temp", email="temp@user.com", password="tempuser"
        )
        expected_deleted_string = f"user {temp_user.pk} (deleted)"

        self.home.save_revision().publish(user=temp_user)
        temp_user.delete()

        response = self.get()
        self.assertContains(response, expected_deleted_string)


class TestAgingPagesViewPermissions(BaseReportViewTestCase):
    url_name = "wagtailadmin_reports:aging_pages"

    def setUp(self):
        self.user = self.login()

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_with_no_permission(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        # No GroupPagePermission created

        response = self.get()
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_minimal_permissions(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        GroupPagePermission.objects.create(
            group=group,
            page=Page.objects.first(),
            permission_type="add",
        )

        response = self.get()

        self.assertEqual(response.status_code, 200)


class TestFilteredAgingPagesView(BaseReportViewTestCase):
    fixtures = ["test.json"]
    url_name = "wagtailadmin_reports:aging_pages"

    def setUp(self):
        self.user = self.login()
        self.home_page = Page.objects.get(slug="home")
        self.aboutus_page = Page.objects.get(slug="about-us")

    def test_filter_by_live(self):
        response = self.get(params={"live": "true"})

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.home_page.title)
        self.assertContains(response, self.aboutus_page.title)

        response = self.get(params={"live": "false"})
        self.assertEqual(response.status_code, 200)
        self.assertNotContains(response, self.home_page.title)
        self.assertNotContains(response, self.aboutus_page.title)

    def test_filter_by_content_type(self):
        ct_pk = self.aboutus_page.specific.content_type.pk
        response = self.get(params={"content_type": ct_pk})

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.aboutus_page.title)
        self.assertNotContains(response, self.home_page.title)

        soup = self.get_soup(response.content)
        self.assertActiveFilter(soup, "content_type", ct_pk)

        # Should render the filter inside the drilldown component
        ct_select = soup.select_one(
            f"{self.drilldown_selector} select[name='content_type']"
        )
        self.assertIsNotNone(ct_select)
        selected_option = ct_select.select_one("option[selected]")
        self.assertIsNotNone(selected_option)
        self.assertEqual(selected_option.get("value"), str(ct_pk))

    def test_filter_by_last_published_at(self):
        self.home_page.last_published_at = timezone.now()
        self.home_page.save()

        response = self.get(params={"last_published_at": "2015-01-01"})
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.aboutus_page.title)
        self.assertNotContains(response, self.home_page.title)


class TestFilteredAgingPagesResultsView(TestFilteredAgingPagesView):
    url_name = "wagtailadmin_reports:aging_pages_results"
    results_only = True


class PageTypesUsageReportViewTest(BaseReportViewTestCase):
    fixtures = ["test.json"]
    url_name = "wagtailadmin_reports:page_types_usage"

    def setUp(self):
        self.user = self.login()

    @staticmethod
    def display_name(content_type):
        return f"{content_type.app_label}.{content_type.model}"

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/reports/base_report.html")
        self.assertTemplateUsed(
            response,
            "wagtailadmin/reports/page_types_usage_results.html",
        )
        self.assertBreadcrumbs(
            [{"url": "", "label": "Page types usage"}],
            response.content,
        )
        soup = self.get_soup(response.content)
        self.assertActiveFilterNotRendered(soup)
        self.assertPageTitle(soup, "Page types usage - Wagtail")

    def test_displays_only_page_types(self):
        """Asserts that the correct models are included in the queryset."""
        response = self.get()
        # Assert that the response contains page models:
        event_page_content_type = ContentType.objects.get_for_model(EventPage)
        event_page_content_type_full_name = self.display_name(event_page_content_type)
        self.assertContains(response, event_page_content_type_full_name)
        simple_page_content_type = ContentType.objects.get_for_model(SimplePage)
        simple_page_content_type_full_name = self.display_name(simple_page_content_type)
        self.assertContains(response, simple_page_content_type_full_name)
        # But it should not contain non-page models:
        event_page_speaker_content_type = ContentType.objects.get_for_model(
            EventPageSpeaker
        )
        event_page_speaker_content_type_full_name = self.display_name(
            event_page_speaker_content_type
        )
        self.assertNotContains(response, event_page_speaker_content_type_full_name)

    def test_displays_wagtailcore_page_if_has_instances(self):
        """Asserts that the wagtailcore.Page model is included in the queryset if it has instances."""
        page_content_type = ContentType.objects.get_for_model(Page)
        page_content_type_full_name = self.display_name(page_content_type)

        # Start with no pages:
        Page.objects.filter(depth__gt=1, content_type=page_content_type).delete()

        # There aren't any Page instances and it is not creatable, it shouldn't be included in the report
        response = self.get()
        self.assertNotContains(response, page_content_type_full_name)

        # Create a page:
        page = Page(title="Page")
        Page.get_first_root_node().add_child(instance=page)

        # There is a Page now, so report should include the `Page` ContentType
        response = self.get()
        self.assertContains(response, page_content_type_full_name)


class PageTypesUsageReportViewQuerysetTests(WagtailTestUtils, TestCase):
    def setUp(self):
        super().setUp()
        self.view = page_types_usage.PageTypesUsageReportView()
        self.view.request = RequestFactory().get(
            reverse("wagtailadmin_reports:page_types_usage")
        )
        self.root = Page.objects.first()
        self.home = Page.objects.get(slug="home")

        self.simple_page_a = SimplePage(title="Simple page A", content="hello")
        self.simple_page_b = SimplePage(title="Simple page B", content="hello")
        self.simple_page_c = SimplePage(title="Simple page C", content="hello")

        Page.get_first_root_node().add_child(instance=self.simple_page_a)
        Page.get_first_root_node().add_child(instance=self.simple_page_b)
        Page.get_first_root_node().add_child(instance=self.simple_page_c)
        self.simple_page_a.save_revision().publish()
        self.simple_page_b.save_revision().publish()
        self.simple_page_c.save_revision().publish()

        self.event_page = EventPage(
            title="Event Page",
            audience="public",
            location="foo",
            cost="bar",
            date_from=datetime.date.today(),
        )
        Page.get_first_root_node().add_child(instance=self.event_page)

    def test_queryset_ordering(self):
        """Asserts that the queryset is ordered by page model count."""
        # Test that the queryset is correctly ordered by page model count
        queryset = self.view.get_queryset()
        # Convert queryset to list of ids to make it easier to test
        queryset_list_pks = list(queryset.values_list("pk", flat=True))
        # Get the positions of the content types in the list
        simple_page_content_type = ContentType.objects.get_for_model(SimplePage)
        simple_page_position = queryset_list_pks.index(simple_page_content_type.pk)
        event_page_content_type = ContentType.objects.get_for_model(EventPage)
        event_page_position = queryset_list_pks.index(event_page_content_type.pk)
        # Assert that the SimplePage comes before EventPage, since it has more entries
        self.assertTrue(simple_page_position < event_page_position)
        # There should be 3 SimplePages
        self.assertEqual(queryset.get(id=simple_page_content_type.pk).count, 3)
        # There should be 1 EventPage
        self.assertEqual(queryset.get(id=event_page_content_type.pk).count, 1)

    def test_queryset_last_edited_page(self):
        """Tests that the queryset correctly returns the last edited page."""
        # Edit the first simple page
        revision = self.simple_page_a.save_revision()
        revision.publish()
        # Edit the second simple page
        revision = self.simple_page_b.save_revision()
        revision.publish()
        # Edit the third simple page
        revision = self.simple_page_c.save_revision()
        revision.publish()
        # Re-edit the first simple page
        revision = self.simple_page_a.save_revision()
        revision.publish()
        # Get the queryset:
        queryset = self.view.decorate_paginated_queryset(self.view.get_queryset())
        # Assert that the first simple page is the last edited page
        self.simple_page_a.refresh_from_db()
        self.assertEqual(queryset[0].last_edited_page.specific, self.simple_page_a)


@override_settings(LANGUAGE_CODE="en", WAGTAIL_I18N_ENABLED=True)
class PageTypesReportFiltersTests(BaseReportViewTestCase):
    url_name = "wagtailadmin_reports:page_types_usage"

    def setUp(self):
        self.user = self.login()
        self.default_locale = Locale.get_default()
        self.fr_locale, _ = Locale.objects.get_or_create(language_code="fr")

    def test_locale_filtering(self):
        # Create pages in default locale
        event_page = EventPage(
            title="Event Page",
            audience="public",
            location="foo",
            cost="bar",
            date_from=datetime.date.today(),
        )
        simple_page = SimplePage(title="Simple Page", content="hello")
        Page.get_first_root_node().add_child(instance=event_page)
        Page.get_first_root_node().add_child(instance=simple_page)
        event_page.save_revision().publish()
        simple_page.save_revision().publish()
        # Translate pages to French
        event_page.copy_for_translation(self.fr_locale)
        simple_page.copy_for_translation(self.fr_locale)

        # Edit the simple page in English to make sure that it's the latest
        simple_page.title = "Updated Simple Page English title"
        revision = simple_page.save_revision()
        simple_page.publish(revision)

        response = self.get()
        page_types = {
            content_type.id: content_type
            for content_type in response.context["object_list"]
        }

        event_page_row = page_types.get(ContentType.objects.get_for_model(EventPage).pk)
        simple_page_row = page_types.get(
            ContentType.objects.get_for_model(SimplePage).pk
        )

        self.assertEqual(event_page_row.count, 2)
        self.assertEqual(simple_page_row.count, 2)
        # The last edited page should be the French version
        self.assertEqual(event_page_row.last_edited_page.locale, self.fr_locale)
        # The last edited SimplePage should be the English version
        self.assertEqual(simple_page_row.last_edited_page.locale, self.default_locale)

        # Filter by French locale
        response = self.get({"page_locale": self.fr_locale.language_code})
        page_types = {
            content_type.id: content_type
            for content_type in response.context["object_list"]
        }

        event_page_row = page_types.get(ContentType.objects.get_for_model(EventPage).pk)
        simple_page_row = page_types.get(
            ContentType.objects.get_for_model(SimplePage).pk
        )

        # There should be 1 of each page (only the French locale ones)
        self.assertEqual(event_page_row.count, 1)
        self.assertEqual(simple_page_row.count, 1)
        # The last edited page should be the French version (even though page was later edited in English)
        self.assertEqual(event_page_row.last_edited_page.locale, self.fr_locale)
        self.assertEqual(simple_page_row.last_edited_page.locale, self.fr_locale)

        # Should render the filter inside the drilldown component
        soup = self.get_soup(response.content)
        locale_select = soup.select_one(
            f"{self.drilldown_selector} select[name='page_locale']"
        )
        self.assertIsNotNone(locale_select)
        selected_option = locale_select.select_one("option[selected]")
        self.assertIsNotNone(selected_option)
        self.assertEqual(selected_option.get("value"), "fr")

    def test_site_filtering_with_single_site(self):
        """Asserts that the site filter is not displayed when there is only one site."""
        sites = Site.objects.all()
        self.assertEqual(sites.count(), 1)

        response = self.get()
        filterset = response.context["filters"]

        # Assert that the filterset does not have the site field
        self.assertNotIn("site", filterset.form.fields)
        self.assertNotIn("site", filterset.filters.keys())
        self.assertFalse(filterset.sites_filter_enabled)

    def test_site_filtering_with_multiple_sites(self):
        root_page = Page.get_first_root_node()
        # Create pages in default locale
        event_page = EventPage(
            title="Event Page",
            audience="public",
            location="foo",
            cost="bar",
            date_from=datetime.date.today(),
        )
        simple_page = SimplePage(title="Simple Page", content="hello")
        root_page.add_child(instance=event_page)
        root_page.add_child(instance=simple_page)

        # Create a new site and add the pages to it
        simple_page_site = Site.objects.create(
            hostname="example.com", root_page=simple_page, is_default_site=False
        )
        self.assertEqual(Site.objects.count(), 2)

        response = self.get()
        page_types = {
            content_type.id: content_type
            for content_type in response.context["object_list"]
        }

        event_page_row = page_types.get(ContentType.objects.get_for_model(EventPage).pk)
        simple_page_row = page_types.get(
            ContentType.objects.get_for_model(SimplePage).pk
        )

        self.assertEqual(event_page_row.count, 1)
        self.assertEqual(simple_page_row.count, 1)

        # Filter by the simple_page_site
        response = self.get({"site": simple_page_site.root_page.path})
        page_types = {
            content_type.id: content_type
            for content_type in response.context["object_list"]
        }

        simple_page_row = page_types.get(
            ContentType.objects.get_for_model(SimplePage).pk
        )

        # There should be 1 SimplePage
        self.assertEqual(simple_page_row.count, 1)
        # There shouldn't be a regular Page
        self.assertFalse(ContentType.objects.get_for_model(EventPage).pk in page_types)

    @override_settings(
        WAGTAIL_CONTENT_LANGUAGES=[
            ("en", "English"),
            ("de", "German"),
            ("fr", "French"),
        ],
    )
    def test_get_locale_choices(self):
        choices = page_types_usage._get_locale_choices()

        expected_choices = [
            ("en", "English"),
            ("de", "German"),
            ("fr", "French"),
        ]

        self.assertCountEqual(choices, expected_choices)


class PageTypesReportFiltersResultsTests(PageTypesReportFiltersTests):
    url_name = "wagtailadmin_reports:page_types_usage_results"
    results_only = True


class TestPageTypesUsageReportViewPermissions(BaseReportViewTestCase):
    fixtures = ["test.json"]
    url_name = "wagtailadmin_reports:page_types_usage"

    def test_simple(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)

    def test_get_with_no_permission(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        # No GroupPagePermission created

        response = self.get()
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_get_with_minimal_permissions(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        GroupPagePermission.objects.create(
            group=group,
            page=Page.objects.first(),
            permission_type="add",
        )

        response = self.get()

        self.assertEqual(response.status_code, 200)

    def test_get_with_page_specific_permissions(self):
        group = Group.objects.create(name="test group")
        self.user.is_superuser = False
        self.user.save()
        self.user.groups.add(group)
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        latest_edited_event_page = EventPage.objects.order_by(
            F("latest_revision_created_at").desc(nulls_last=True), "title", "-pk"
        ).first()
        latest_edited_simple_page = SimplePage.objects.order_by(
            F("latest_revision_created_at").desc(nulls_last=True), "title", "-pk"
        ).first()
        GroupPagePermission.objects.create(
            group=group,
            page=latest_edited_event_page,
            permission_type="change",
        )

        response = self.get()

        self.assertEqual(response.status_code, 200)
        # For pages that the user can edit, it should show the page title and link to edit the page:
        edit_event_page_url = reverse(
            "wagtailadmin_pages:edit", args=(latest_edited_event_page.id,)
        )
        self.assertContains(
            response,
            f"<a href={edit_event_page_url}>{latest_edited_event_page.get_admin_display_title()}</a>",
            html=True,
        )
        # For pages that the user cannot edit, it should only show the page title
        self.assertContains(
            response,
            f"<p>{latest_edited_simple_page.get_admin_display_title()}</p>",
            html=True,
        )
        edit_simple_page_url = reverse(
            "wagtailadmin_pages:edit", args=(latest_edited_simple_page.id,)
        )
        self.assertNotContains(
            response,
            f"<a href={edit_simple_page_url}>{latest_edited_simple_page.get_admin_display_title()}</a>",
            html=True,
        )



from django.contrib.auth.models import Group, Permission
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from django.urls import reverse

from wagtail.models import GroupPagePermission, Page
from wagtail.test.testapp.models import (
    CustomCopyFormPage,
    EventPage,
    EventPageSpeaker,
    PageWithExcludedCopyField,
    SimplePage,
)
from wagtail.test.utils import WagtailTestUtils


class TestPageCopy(WagtailTestUtils, TestCase):
    def setUp(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)

        # Create a page
        self.test_page = self.root_page.add_child(
            instance=SimplePage(
                title="Hello world!",
                slug="hello-world",
                content="hello",
                live=True,
                has_unpublished_changes=False,
            )
        )

        # Create a couple of child pages
        self.test_child_page = self.test_page.add_child(
            instance=SimplePage(
                title="Child page",
                slug="child-page",
                content="hello",
                live=True,
                has_unpublished_changes=True,
            )
        )

        self.test_unpublished_child_page = self.test_page.add_child(
            instance=SimplePage(
                title="Unpublished Child page",
                slug="unpublished-child-page",
                content="hello",
                live=False,
                has_unpublished_changes=True,
            )
        )

        # Login
        self.user = self.login()

    def test_page_copy(self):
        response = self.client.get(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,))
        )

        # Check response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/pages/copy.html")

        # Make sure all fields are in the form
        self.assertContains(response, "New title")
        self.assertContains(response, "New slug")
        self.assertContains(response, "New parent page")
        self.assertContains(response, "Copy subpages")
        self.assertContains(response, "Publish copies")
        self.assertContains(response, "Alias")

    def test_page_copy_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        # Get copy page
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world",
            "new_parent_page": str(self.test_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # A user with no page permissions at all should be redirected to the admin home
        self.assertRedirects(response, reverse("wagtailadmin_home"))

        # A user with page permissions, but not add permission at the destination,
        # should receive a form validation error
        publishers = Group.objects.create(name="Publishers")
        GroupPagePermission.objects.create(
            group=publishers, page=self.root_page, permission_type="publish"
        )
        self.user.groups.add(publishers)
        self.user.save()

        # Get copy page
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world",
            "new_parent_page": str(self.test_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )
        form = response.context["form"]
        self.assertFalse(form.is_valid())
        self.assertIn("new_parent_page", form.errors)

    def test_page_copy_post(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()

        # Check that the copy exists
        self.assertIsNotNone(page_copy)

        # Check that the copy is not live
        self.assertFalse(page_copy.live)
        self.assertTrue(page_copy.has_unpublished_changes)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were not copied
        self.assertEqual(page_copy.get_children().count(), 0)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_with_exclude_fields_in_copy(self):
        original_page = self.test_page.add_child(
            instance=PageWithExcludedCopyField(
                title="Page with exclude_fields_in_copy",
                slug="page-with-exclude-fields-in-copy",
                content="Copy me",
                special_field="Don't copy me",
                live=True,
                has_unpublished_changes=False,
            )
        )
        post_data = {
            "new_title": f"{original_page.title} 2",
            "new_slug": f"{original_page.slug}-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        self.client.post(
            reverse("wagtailadmin_pages:copy", args=(original_page.id,)), post_data
        )
        # Get copy
        page_copy = PageWithExcludedCopyField.objects.get(slug=post_data["new_slug"])
        self.assertEqual(page_copy.content, original_page.content)
        self.assertNotEqual(page_copy.special_field, original_page.special_field)
        self.assertEqual(
            page_copy.special_field, page_copy._meta.get_field("special_field").default
        )

    def test_page_copy_post_copy_subpages(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": True,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()

        # Check that the copy exists
        self.assertIsNotNone(page_copy)

        # Check that the copy is not live
        self.assertFalse(page_copy.live)
        self.assertTrue(page_copy.has_unpublished_changes)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were copied
        self.assertEqual(page_copy.get_children().count(), 2)

        # Check the child pages
        # Neither of them should be live
        child_copy = page_copy.get_children().filter(slug="child-page").first()
        self.assertIsNotNone(child_copy)
        self.assertFalse(child_copy.live)
        self.assertTrue(child_copy.has_unpublished_changes)

        unpublished_child_copy = (
            page_copy.get_children().filter(slug="unpublished-child-page").first()
        )
        self.assertIsNotNone(unpublished_child_copy)
        self.assertFalse(unpublished_child_copy.live)
        self.assertTrue(unpublished_child_copy.has_unpublished_changes)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_copy_post_copy_subpages_publish_copies(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": True,
            "publish_copies": True,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()

        # Check that the copy exists
        self.assertIsNotNone(page_copy)

        # Check that the copy is live
        self.assertTrue(page_copy.live)
        self.assertFalse(page_copy.has_unpublished_changes)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were copied
        self.assertEqual(page_copy.get_children().count(), 2)

        # Check the child pages
        # The child_copy should be live but the unpublished_child_copy shouldn't
        child_copy = page_copy.get_children().filter(slug="child-page").first()
        self.assertIsNotNone(child_copy)
        self.assertTrue(child_copy.live)
        self.assertTrue(child_copy.has_unpublished_changes)

        unpublished_child_copy = (
            page_copy.get_children().filter(slug="unpublished-child-page").first()
        )
        self.assertIsNotNone(unpublished_child_copy)
        self.assertFalse(unpublished_child_copy.live)
        self.assertTrue(unpublished_child_copy.has_unpublished_changes)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_copy_post_new_parent(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.test_child_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the new parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.test_child_page.id,))
        )

        # Check that the page was copied to the correct place
        self.assertTrue(
            Page.objects.filter(slug="hello-world-2").first().get_parent(),
            msg=self.test_child_page,
        )

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_copy_post_existing_slug_within_same_parent_page(self):
        # This tests the existing slug checking on page copy when not changing the parent page

        # Attempt to copy the page but forget to change the slug
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Should not be redirected (as the save should fail)
        self.assertEqual(response.status_code, 200)

        # Check that a form error was raised
        self.assertFormError(
            response.context["form"],
            "new_slug",
            'This slug is already in use within the context of its parent page "Welcome to your new Wagtail site!"',
        )

    def test_page_copy_post_and_subpages_to_same_tree_branch(self):
        # This tests that a page cannot be copied into itself when copying subpages
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world",
            "new_parent_page": str(self.test_child_page.id),
            "copy_subpages": True,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Should not be redirected (as the save should fail)
        self.assertEqual(response.status_code, 200)

        # Check that a form error was raised
        self.assertFormError(
            response.context["form"],
            "new_parent_page",
            "You cannot copy a page into itself when copying subpages",
        )

    def test_page_copy_post_existing_slug_to_another_parent_page(self):
        # This tests the existing slug checking on page copy when changing the parent page

        # Attempt to copy the page and changed the parent page
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world",
            "new_parent_page": str(self.test_child_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.test_child_page.id,))
        )

    def test_page_copy_post_invalid_slug(self):
        # Attempt to copy the page but set an invalid slug string
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello world!",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Should not be redirected (as the save should fail)
        self.assertEqual(response.status_code, 200)

        # Check that a form error was raised
        self.assertFormError(
            response.context["form"],
            "new_slug",
            "Enter a valid “slug” consisting of Unicode letters, numbers, underscores, or hyphens.",
        )

    def test_page_copy_post_valid_unicode_slug(self):
        post_data = {
            "new_title": "Hello wɜːld",
            "new_slug": "hello-wɜːld",
            "new_parent_page": str(self.test_page.id),
            "copy_subpages": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check response
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.test_page.id,))
        )

        # Get copy
        page_copy = (
            self.test_page.get_children().filter(slug=post_data["new_slug"]).first()
        )

        # Check that the copy exists with the good slug
        self.assertIsNotNone(page_copy)
        self.assertEqual(page_copy.slug, post_data["new_slug"])

    def test_page_copy_no_publish_permission(self):
        # Turn user into an editor who can add pages but not publish them
        self.user.is_superuser = False
        self.user.groups.add(
            Group.objects.get(name="Editors"),
        )
        self.user.save()

        # Get copy page
        response = self.client.get(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,))
        )

        # The user should have access to the copy page
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/pages/copy.html")

        # Make sure the "publish copies" field is hidden
        self.assertNotContains(response, "Publish copies")

    def test_page_copy_no_publish_permission_post_copy_subpages_publish_copies(self):
        # This tests that unprivileged users cannot publish copied pages even if they hack their browser

        # Turn user into an editor who can add pages but not publish them
        self.user.is_superuser = False
        self.user.groups.add(
            Group.objects.get(name="Editors"),
        )
        self.user.save()

        # Post
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": True,
            "publish_copies": True,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()

        # Check that the copy exists
        self.assertIsNotNone(page_copy)

        # Check that the copy is not live
        self.assertFalse(page_copy.live)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were copied
        self.assertEqual(page_copy.get_children().count(), 2)

        # Check the child pages
        # Neither of them should be live
        child_copy = page_copy.get_children().filter(slug="child-page").first()
        self.assertIsNotNone(child_copy)
        self.assertFalse(child_copy.live)

        unpublished_child_copy = (
            page_copy.get_children().filter(slug="unpublished-child-page").first()
        )
        self.assertIsNotNone(unpublished_child_copy)
        self.assertFalse(unpublished_child_copy.live)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_before_copy_page_hook(self):
        def hook_func(request, page):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("before_copy_page", hook_func):
            response = self.client.get(
                reverse("wagtailadmin_pages:copy", args=(self.test_page.id,))
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

    def test_before_copy_page_hook_post(self):
        def hook_func(request, page):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("before_copy_page", hook_func):
            post_data = {
                "new_title": "Hello world 2",
                "new_slug": "hello-world-2",
                "new_parent_page": str(self.root_page.id),
                "copy_subpages": False,
                "publish_copies": False,
                "alias": False,
            }
            response = self.client.post(
                reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # page should not be copied
        self.assertFalse(Page.objects.filter(title="Hello world 2").exists())

    def test_after_copy_page_hook(self):
        def hook_func(request, page, new_page):
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(page.specific, SimplePage)
            self.assertIsInstance(new_page.specific, SimplePage)

            return HttpResponse("Overridden!")

        with self.register_hook("after_copy_page", hook_func):
            post_data = {
                "new_title": "Hello world 2",
                "new_slug": "hello-world-2",
                "new_parent_page": str(self.root_page.id),
                "copy_subpages": False,
                "publish_copies": False,
                "alias": False,
            }
            response = self.client.post(
                reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # page should be copied
        self.assertTrue(Page.objects.filter(title="Hello world 2").exists())

    def test_page_copy_alias_post(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": True,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().get(slug="hello-world-2")

        # Check the copy is an alias of the original
        self.assertEqual(page_copy.alias_of, self.test_page.page_ptr)

        # Check that the copy is live
        # Note: publish_copies is ignored. Alias pages always keep the same state as their original
        self.assertTrue(page_copy.live)
        self.assertFalse(page_copy.has_unpublished_changes)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were not copied
        self.assertEqual(page_copy.get_children().count(), 0)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_copy_alias_post_copy_subpages(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": True,
            "publish_copies": False,
            "alias": True,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)), post_data
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().get(slug="hello-world-2")

        # Check the copy is an alias of the original
        self.assertEqual(page_copy.alias_of, self.test_page.page_ptr)

        # Check that the copy is live
        # Note: publish_copies is ignored. Alias pages always keep the same state as their original
        self.assertTrue(page_copy.live)
        self.assertFalse(page_copy.has_unpublished_changes)

        # Check that the owner of the page is set correctly
        self.assertEqual(page_copy.owner, self.user)

        # Check that the children were copied
        self.assertEqual(page_copy.get_children().count(), 2)

        # Check the child pages
        # Neither of them should be live
        child_copy = page_copy.get_children().filter(slug="child-page").first()
        self.assertIsNotNone(child_copy)
        self.assertEqual(child_copy.alias_of, self.test_child_page.page_ptr)
        self.assertTrue(child_copy.live)
        self.assertFalse(child_copy.has_unpublished_changes)

        unpublished_child_copy = (
            page_copy.get_children().filter(slug="unpublished-child-page").first()
        )
        self.assertIsNotNone(unpublished_child_copy)
        self.assertEqual(
            unpublished_child_copy.alias_of, self.test_unpublished_child_page.page_ptr
        )
        self.assertFalse(unpublished_child_copy.live)
        self.assertTrue(unpublished_child_copy.has_unpublished_changes)

        # treebeard should report no consistency problems with the tree
        self.assertFalse(
            any(Page.find_problems()), msg="treebeard found consistency problems"
        )

    def test_page_copy_alias_post_without_source_publish_permission(self):
        # Check for issue #7293 - If the user has permission to publish at a destination, but not the source.
        # Wagtail would crash on attempt to copy

        # Create a new section
        self.destination_page = self.root_page.add_child(
            instance=SimplePage(
                title="Destination page",
                slug="destination-page",
                content="hello",
                live=True,
                has_unpublished_changes=False,
            )
        )

        # Make user a moderator and make it so they can only publish at the destination page
        self.user.is_superuser = False
        self.user.groups.add(Group.objects.get(name="Moderators"))
        self.user.save()
        GroupPagePermission.objects.filter(permission__codename="publish_page").update(
            page=self.destination_page
        )

        post_data = {
            "new_title": self.test_child_page.title,
            "new_slug": self.test_child_page.slug,
            "new_parent_page": str(self.destination_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=[self.test_child_page.id]),
            post_data,
        )

        # We only need to check that it didn't crash
        self.assertEqual(response.status_code, 302)

    def test_copy_page_with_unique_uuids_in_orderables(self):
        """
        Test that a page with orderables can be copied and the translation
        keys are updated.
        """
        event_page = EventPage(
            title="Moon Landing",
            location="the moon",
            audience="public",
            cost="free on TV",
            date_from="1969-07-20",
        )
        self.root_page.add_child(instance=event_page)

        event_page.speakers.add(
            EventPageSpeaker(
                first_name="Neil",
                last_name="Armstrong",
            )
        )
        # ensure there's a revision (which should capture the new speaker orderables)
        # before copying the page
        event_page.save_revision().publish()

        post_data = {
            "new_title": "New Moon landing",
            "new_slug": "moon-landing-redux",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        self.client.post(
            reverse("wagtailadmin_pages:copy", args=[event_page.id]), post_data
        )
        new_page = EventPage.objects.last()

        # Hack: get the page instance from the edit form which assembles it from the
        # latest revision. While we could do new_page.get_latest_revision().as_object()
        # this follow the edit view closer and should it change the test is less
        # prone to continue working because we're skipping some step
        response = self.client.get(
            reverse("wagtailadmin_pages:edit", args=[new_page.id])
        )
        new_page_on_edit_form = response.context["form"].instance

        # publish the page, similar to EditView.publish_action()
        new_page_on_edit_form.save_revision().publish()

        self.assertNotEqual(
            event_page.speakers.first().translation_key,
            new_page.speakers.first().translation_key,
        )

    def test_page_copy_with_custom_copy_form(self):
        custom_copy_form_page = self.root_page.add_child(
            instance=CustomCopyFormPage(
                title="Hello world!",
                slug="copy-form",
                live=True,
                has_unpublished_changes=False,
            )
        )

        response = self.client.get(
            reverse("wagtailadmin_pages:copy", args=(custom_copy_form_page.id,))
        )
        # Check response
        self.assertEqual(response.status_code, 200)

        # Check if slug is hello-world-2
        self.assertContains(response, "copy-form-2")

    def test_copy_page_success_msg_contains_edit_button(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": False,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)),
            post_data,
            follow=True,
        )

        self.assertEqual(response.status_code, 200)

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()
        self.assertIsNotNone(page_copy)

        # Check the "Edit" button present in success message
        soup = self.get_soup(response.content)
        edit_button = soup.select_one(".messages .success a.button")
        self.assertIsNotNone(edit_button)
        self.assertEqual(
            edit_button["href"],
            reverse("wagtailadmin_pages:edit", args=(page_copy.id,)),
        )
        self.assertEqual("Edit", edit_button.text)

    def test_copy_page_success_msg_contains_edit_button_subpages(self):
        post_data = {
            "new_title": "Hello world 2",
            "new_slug": "hello-world-2",
            "new_parent_page": str(self.root_page.id),
            "copy_subpages": True,
            "publish_copies": False,
            "alias": False,
        }
        response = self.client.post(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,)),
            post_data,
            follow=True,
        )

        # Check that the user was redirected to the parents explore page
        self.assertRedirects(
            response, reverse("wagtailadmin_explore", args=(self.root_page.id,))
        )

        # Get copy
        page_copy = self.root_page.get_children().filter(slug="hello-world-2").first()
        self.assertIsNotNone(page_copy)

        # Check the "Edit" button present in success message
        soup = self.get_soup(response.content)
        edit_button = soup.select_one(".messages .success a.button")
        self.assertIsNotNone(edit_button)
        self.assertEqual(
            edit_button["href"],
            reverse("wagtailadmin_pages:edit", args=(page_copy.id,)),
        )
        self.assertEqual("Edit", edit_button.text)

    def test_page_copy_form_widgets(self):
        """
        For widget rendering, we want to validate that the locale gets passed to the slug
        and that the title is NOT syncing with the slug as this is NOT required on copy form.
        """

        response = self.client.get(
            reverse("wagtailadmin_pages:copy", args=(self.test_page.id,))
        )

        # Check response
        self.assertEqual(response.status_code, 200)

        # Check that the form has the correct widgets
        self.assertContains(response, "id_new_title")
        self.assertContains(response, "id_new_slug")

        soup = self.get_soup(response.content)

        # Check the attributes on the slug field (locale available & slugify used on blue)
        slug_field = soup.find("input", id="id_new_slug")
        self.assertEqual(
            slug_field.attrs,
            {
                "type": "text",
                "name": "new_slug",
                "value": "hello-world",
                "data-controller": "w-slug",
                "data-action": "blur->w-slug#slugify w-sync:check->w-slug#compare w-sync:apply->w-slug#urlify:prevent",
                "data-w-slug-allow-unicode-value": "",
                "data-w-slug-compare-as-param": "urlify",
                "data-w-slug-trim-value": "true",
                "data-w-slug-locale-value": "en",
                "required": "",
                "id": "id_new_slug",
            },
        )

        # Check the attributes on the title field (no syncing)
        title_field = soup.find("input", id="id_new_title")
        self.assertEqual(
            title_field.attrs,
            {
                "type": "text",
                "name": "new_title",
                "value": "Hello world!",
                "required": "",
                "id": "id_new_title",
            },
        )



# Generated by Django 4.0.3 on 2022-05-06 04:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0075_populate_latest_revision_and_revision_object_str"),
    ]

    operations = [
        migrations.AddField(
            model_name="modellogentry",
            name="revision",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wagtailcore.revision",
            ),
        ),
    ]



from django.urls import path

from wagtail.embeds.views import chooser

app_name = "wagtailembeds"
urlpatterns = [
    path("chooser/", chooser.chooser, name="chooser"),
    path("chooser/upload/", chooser.chooser_upload, name="chooser_upload"),
]



from itertools import zip_longest

from factory import SubFactory
from factory.builder import StepBuilder

from wagtail import blocks


class StreamFieldFactoryException(Exception):
    pass


class InvalidDeclaration(StreamFieldFactoryException):
    pass


class DuplicateDeclaration(StreamFieldFactoryException):
    pass


class UnknownChildBlockFactory(StreamFieldFactoryException):
    pass


class BaseBlockStepBuilder(StepBuilder):
    def recurse(self, factory_meta, extras):
        """Recurse into a sub-factory call."""
        builder_class = factory_meta.factory._builder_class
        return builder_class(factory_meta, extras, strategy=self.strategy)


class StructBlockStepBuilder(BaseBlockStepBuilder):
    pass


class ListBlockStepBuilder(BaseBlockStepBuilder):
    pass


class StreamBlockStepBuilder(BaseBlockStepBuilder):
    def __init__(self, factory_meta, extras, strategy):
        indexed_block_names, extra_declarations = self.get_block_declarations(
            factory_meta, extras
        )
        new_factory_class = self.create_factory_class(factory_meta, indexed_block_names)
        super().__init__(new_factory_class._meta, extra_declarations, strategy)

    def get_block_declarations(self, factory_meta, extras):
        # Mapping of StreamValue index -> block name. We will use this to create a
        # StreamBlockFactory subclass with one declaration for each pair, named
        # <index>.<block_name>
        indexed_block_names = {}

        # Extra declarations passed at instantiation, renamed from <index>__<name>__... to
        # <index>.<block_name>__..., to match the declarations on the StreamBlockFactory subclass
        # we will generate. As DeclarationSet splits parameter names on "__" the
        # <index>.<block_name> keys won't cause errors for unknown declarations (0__foo_block
        # implies a declaration "0" with context "foo_block"). They will also have the important
        # property of being uniquely hashable
        extra_declarations = {}

        for k, v in extras.items():
            if k.isdigit():
                # We got a declaration like `<index>="foo_block"' - <index> should get the
                # default value for foo_block, so don't store this item in extra_declarations
                if v not in factory_meta.base_declarations:
                    raise UnknownChildBlockFactory(
                        f"No factory defined for block '{v}'"
                    )
                key = int(k)
                if key in indexed_block_names and indexed_block_names[key] != v:
                    raise DuplicateDeclaration(
                        f"Multiple declarations for index {key} at this level of nesting "
                        f"(got {v}, already have {indexed_block_names[key]})"
                    )
                indexed_block_names[key] = v
            else:
                try:
                    i, name, *params = k.split("__", maxsplit=2)
                    key = int(i)
                except (ValueError, TypeError):
                    raise InvalidDeclaration(
                        "StreamFieldFactory declarations must be of the form "
                        "<index>=<block_name>, <index>__<block_name>=value or "
                        f"<index>__<block_name>__<param>=value, got: {k}"
                    )
                if key in indexed_block_names and indexed_block_names[key] != name:
                    raise DuplicateDeclaration(
                        f"Multiple declarations for index {key} at this level of nesting "
                        f"(got {name}, already have {indexed_block_names[key]})"
                    )
                indexed_block_names[key] = name
                transformed_key = self.reconstruct_key(i, name, params)
                extra_declarations[transformed_key] = v

        self.validate_block_indexes_sequential(indexed_block_names, factory_meta)
        return indexed_block_names, extra_declarations

    def reconstruct_key(self, index, name, params):
        return f"{index}.{'__'.join((name, *params))}"

    def validate_block_indexes_sequential(self, indexed_block_names, factory_meta):
        if not indexed_block_names:
            # There were no declarations for this block, we will ultimately return an empty
            # StreamValue
            return
        indexes = sorted(indexed_block_names.keys())
        for declared, expected in zip_longest(indexes, range(max(indexes) + 1)):
            if declared != expected:
                raise InvalidDeclaration(
                    f"Parameters for {factory_meta.factory} missing required index {expected}"
                )

    def create_factory_class(self, old_factory_meta, indexed_block_names):
        # Create a new StreamBlockFactory subclass, with a declaration for each block the user
        # requested at instantiation. This way we can rely on the factory_boy internals for
        # object generation
        new_class_dict = {"Meta": old_factory_meta.to_meta_class()}

        block_def = old_factory_meta.get_block_definition()
        for i, name in indexed_block_names.items():
            declared_value = old_factory_meta.base_declarations[name]
            if block_def is not None and isinstance(declared_value, SubFactory):
                # Annotate the subfactory's factory with the correct block definition for that
                # branch of the tree, so we can construct a StreamValue if there's no explicit
                # block class defined (e.g. if a nested StreamBlock was declared inline like
                # `inner_stream = StreamBlock(...))'
                child_def = block_def.child_blocks[name]
                if isinstance(child_def, blocks.ListBlock):
                    # ListBlock is a special case as it is a concrete node in the stream block
                    # tree, but ListBlockFactory is a SubFactory subclass, making it "abstract"
                    # in the factory tree
                    child_def = child_def.child_block
                declared_value.get_factory()._meta.block_def = child_def
            new_class_dict[f"{i}.{name}"] = declared_value

        from .blocks import StreamBlockFactory

        return type(
            "_GeneratedStreamBlockFactory", (StreamBlockFactory,), new_class_dict
        )



from django.contrib.auth import get_user_model
from django.db.models import Q

from wagtail.admin.views.bulk_action import BulkAction
from wagtail.admin.views.generic.permissions import PermissionCheckedMixin
from wagtail.permission_policies import ModelPermissionPolicy
from wagtail.search.backends import get_search_backend
from wagtail.search.index import class_is_indexed

User = get_user_model()


class UserBulkAction(PermissionCheckedMixin, BulkAction):
    models = [User]
    permission_policy = ModelPermissionPolicy(User)
    any_permission_required = ["add", "change", "delete"]

    def get_all_objects_in_listing_query(self, parent_id):
        listing_objects = self.model.objects.all().values_list("pk", flat=True)

        if q := self.request.GET.get("q"):
            if class_is_indexed(self.model):
                search_backend = get_search_backend()
                return search_backend.autocomplete(q, listing_objects)

            model_fields = {f.name for f in self.model._meta.get_fields()}
            filterable_fields = {"username", "first_name", "last_name", "email"}
            common_fields = model_fields & filterable_fields
            conditions = Q()
            for field in common_fields:
                conditions |= Q(**{f"{field}__icontains": q})
            return listing_objects.filter(conditions)

        return listing_objects



"""An alternative urlconf module where Wagtail does not serve front-end URLs"""

from django.urls import include, path

from wagtail.admin import urls as wagtailadmin_urls
from wagtail.documents import urls as wagtaildocs_urls
from wagtail.images import urls as wagtailimages_urls

urlpatterns = [
    path("admin/", include(wagtailadmin_urls)),
    path("documents/", include(wagtaildocs_urls)),
    path("images/", include(wagtailimages_urls)),
]



from django.test import SimpleTestCase

from wagtail.admin.icons import get_icon_sprite_hash, get_icon_sprite_url


class TestIconSpriteView(SimpleTestCase):
    def test_content_type(self):
        response = self.client.get(get_icon_sprite_url())
        self.assertEqual(
            response.headers["Content-Type"], "image/svg+xml; charset=utf-8"
        )
        self.assertEqual(response.wsgi_request.GET["h"], get_icon_sprite_hash())

    def test_no_comments(self):
        response = self.client.get(get_icon_sprite_url())
        self.assertNotContains(response, "<!--")


class TestIconSpriteHash(SimpleTestCase):
    def test_hash(self):
        self.assertEqual(len(get_icon_sprite_hash()), 8)



"""Helper classes for formatting data as tables"""

from collections import OrderedDict
from collections.abc import Mapping

from django.contrib.admin.utils import quote
from django.contrib.humanize.templatetags.humanize import intcomma
from django.forms import MediaDefiningClass
from django.template.loader import get_template
from django.templatetags.l10n import unlocalize
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext, gettext_lazy

from wagtail.admin.ui.components import Component
from wagtail.coreutils import get_locales_display_names, multigetattr
from wagtail.models import Locale


class BaseColumn(metaclass=MediaDefiningClass):
    class Header:
        # Helper object used for rendering column headers in templates -
        # behaves as a component (i.e. it has a render_html method) but delegates rendering
        # to Column.render_header_html
        def __init__(self, column):
            self.column = column

        def render_html(self, parent_context):
            return self.column.render_header_html(parent_context)

    class Cell:
        # Helper object used for rendering table cells in templates -
        # behaves as a component (i.e. it has a render_html method) but delegates rendering
        # to Column.render_cell_html
        def __init__(self, column, instance):
            self.column = column
            self.instance = instance

        def render_html(self, parent_context):
            return self.column.render_cell_html(self.instance, parent_context)

    header_template_name = "wagtailadmin/tables/column_header.html"
    cell_template_name = None

    def __init__(
        self,
        name,
        label=None,
        accessor=None,
        classname=None,
        sort_key=None,
        width=None,
        ascending_title_text=None,
        descending_title_text=None,
    ):
        self.name = name
        self.accessor = accessor or name
        if label is None:
            self.label = capfirst(name.replace("_", " "))
        else:
            self.label = label
        self.classname = classname
        self.sort_key = sort_key
        self.header = Column.Header(self)
        self.width = width
        self.ascending_title_text = ascending_title_text
        self.descending_title_text = descending_title_text

    def get_header_context_data(self, parent_context):
        """
        Compiles the context dictionary to pass to the header template when rendering the column header
        """
        table = parent_context["table"]
        return {
            "column": self,
            "table": table,
            "is_orderable": bool(self.sort_key),
            "is_ascending": self.sort_key and table.ordering == self.sort_key,
            "is_descending": self.sort_key and table.ordering == ("-" + self.sort_key),
            "request": parent_context.get("request"),
            "ascending_title_text": self.ascending_title_text
            or table.get_ascending_title_text(self),
            "descending_title_text": self.descending_title_text
            or table.get_descending_title_text(self),
        }

    @cached_property
    def header_template(self):
        return get_template(self.header_template_name)

    @cached_property
    def cell_template(self):
        if self.cell_template_name is None:
            raise NotImplementedError(
                "cell_template_name must be specified on %r" % self
            )
        return get_template(self.cell_template_name)

    def render_header_html(self, parent_context):
        """
        Renders the column's header
        """
        context = self.get_header_context_data(parent_context)
        return self.header_template.render(context)

    def get_cell_context_data(self, instance, parent_context):
        """
        Compiles the context dictionary to pass to the cell template when rendering a table cell for
        the given instance
        """
        return {
            "instance": instance,
            "column": self,
            "row": parent_context["row"],
            "table": parent_context["table"],
            "request": parent_context.get("request"),
        }

    def render_cell_html(self, instance, parent_context):
        """
        Renders a table cell containing data for the given instance
        """
        context = self.get_cell_context_data(instance, parent_context)
        return self.cell_template.render(context)

    def get_cell(self, instance):
        """
        Return an object encapsulating this column and an item of data, which we can use for
        rendering a table cell into a template
        """
        return Column.Cell(self, instance)

    def __repr__(self):
        return "<{}.{}: {}>".format(
            self.__class__.__module__,
            self.__class__.__qualname__,
            self.name,
        )


class Column(BaseColumn):
    """A column that displays a single field of data from the model"""

    cell_template_name = "wagtailadmin/tables/cell.html"
    empty_value_display = ""

    def get_value(self, instance):
        """
        Given an instance (i.e. any object containing data), extract the field of data to be
        displayed in a cell of this column
        """
        if callable(self.accessor):
            return self.accessor(instance)
        else:
            try:
                return multigetattr(instance, self.accessor)
            except AttributeError:
                return None

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["raw_value"] = value = self.get_value(instance)
        if isinstance(value, int) and not isinstance(value, bool):
            # To prevent errors arising from USE_THOUSAND_SEPARATOR, we require all numbers output
            # on templates to be explicitly localized or unlocalized. For numeric table cells, we
            # unlocalize them by default; developers may subclass Column to obtain formatted numbers.
            value = unlocalize(value)

        if not str(value).strip():
            context["value"] = self.empty_value_display
        else:
            context["value"] = value
        return context


class NumberColumn(Column):
    """A specialised column that displays numbers with locale-aware formatting"""

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["value"] = intcomma(context["raw_value"])
        return context


class ButtonsColumnMixin:
    """A mixin for columns that contain buttons"""

    buttons = []

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["buttons"] = sorted(self.get_buttons(instance, parent_context))
        return context

    def get_buttons(self, instance, parent_context):
        return self.buttons


class TitleColumn(Column):
    """A column where data is styled as a title and wrapped in a link or <label>"""

    cell_template_name = "wagtailadmin/tables/title_cell.html"
    empty_value_display = gettext_lazy("(blank)")

    def __init__(
        self,
        name,
        url_name=None,
        get_url=None,
        get_title_id=None,
        label_prefix=None,
        get_label_id=None,
        link_classname=None,
        link_attrs=None,
        id_accessor="pk",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.url_name = url_name
        self._get_url_func = get_url
        self._get_title_id_func = get_title_id
        self.label_prefix = label_prefix
        self._get_label_id_func = get_label_id
        self.link_attrs = link_attrs or {}
        self.link_classname = link_classname
        self.id_accessor = id_accessor

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["link_attrs"] = self.get_link_attrs(instance, parent_context)
        context["link_attrs"]["href"] = context["link_url"] = self.get_link_url(
            instance, parent_context
        )
        if self.link_classname is not None:
            context["link_attrs"]["class"] = self.link_classname
        context["title_id"] = self.get_title_id(instance, parent_context)
        context["label_id"] = self.get_label_id(instance, parent_context)
        return context

    def get_link_attrs(self, instance, parent_context):
        return self.link_attrs.copy()

    def get_link_url(self, instance, parent_context):
        if self._get_url_func:
            return self._get_url_func(instance)
        elif self.url_name:
            id = multigetattr(instance, self.id_accessor)
            return reverse(self.url_name, args=(quote(id),))

    def get_title_id(self, instance, parent_context):
        if self._get_title_id_func:
            return self._get_title_id_func(instance)

    def get_label_id(self, instance, parent_context):
        if self._get_label_id_func:
            return self._get_label_id_func(instance)
        elif self.label_prefix:
            id = multigetattr(instance, self.id_accessor)
            return f"{self.label_prefix}-{id}"


class StatusFlagColumn(Column):
    """Represents a boolean value as a status tag (or lack thereof, if the corresponding label is None)"""

    cell_template_name = "wagtailadmin/tables/status_flag_cell.html"

    def __init__(self, name, true_label=None, false_label=None, **kwargs):
        super().__init__(name, **kwargs)
        self.true_label = true_label
        self.false_label = false_label


class StatusTagColumn(Column):
    """Represents a status tag"""

    cell_template_name = "wagtailadmin/tables/status_tag_cell.html"

    def __init__(self, name, primary=None, **kwargs):
        super().__init__(name, **kwargs)
        self.primary = primary

    def get_primary(self, instance):
        if callable(self.primary):
            return self.primary(instance)
        return self.primary

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["primary"] = self.get_primary(instance)
        return context


class BooleanColumn(Column):
    """Represents a True/False/None value as a tick/cross/question icon"""

    cell_template_name = "wagtailadmin/tables/boolean_cell.html"

    def get_value(self, instance):
        value = super().get_value(instance)
        if value is None:
            return None
        return bool(value)


class LiveStatusTagColumn(StatusTagColumn):
    """Represents a live/draft status tag"""

    def __init__(self, **kwargs):
        super().__init__(
            "status_string",
            label=kwargs.pop("label", gettext("Status")),
            sort_key=kwargs.pop("sort_key", "live"),
            primary=lambda instance: instance.live,
            **kwargs,
        )


class LocaleColumn(Column):
    """Represents a Locale label."""

    cell_template_name = "wagtailadmin/tables/locale_cell.html"

    def __init__(self, **kwargs):
        super().__init__(
            "locale_id",
            label=kwargs.pop("label", gettext("Locale")),
            sort_key=kwargs.pop("sort_key", "locale"),
            classname=kwargs.pop("classname", "w-text-16"),
            **kwargs,
        )

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        value = self.get_value(instance)
        if isinstance(value, int):
            value = get_locales_display_names().get(value)
        elif isinstance(value, Locale):
            value = value.get_display_name()
        context["value"] = value
        return context


class DateColumn(Column):
    """Outputs a date in human-readable format"""

    cell_template_name = "wagtailadmin/tables/date_cell.html"


class UpdatedAtColumn(DateColumn):
    """Outputs the _updated_at date annotation in human-readable format"""

    def __init__(self, **kwargs):
        super().__init__(
            "_updated_at",
            label=kwargs.pop("label", gettext("Updated")),
            sort_key=kwargs.pop("sort_key", "_updated_at"),
            **kwargs,
        )


class UserColumn(Column):
    """Outputs the username and avatar for a user"""

    cell_template_name = "wagtailadmin/tables/user_cell.html"

    def __init__(self, name, blank_display_name="", **kwargs):
        super().__init__(name, **kwargs)
        self.blank_display_name = blank_display_name

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)

        user = context["value"]
        if user:
            try:
                full_name = user.get_full_name().strip()
            except AttributeError:
                full_name = ""
            context["display_name"] = full_name or user.get_username()
        else:
            context["display_name"] = self.blank_display_name
        return context


class BulkActionsCheckboxColumn(BaseColumn):
    """
    A checkbox column for the bulk actions feature.

    When using this column, there should be another column (e.g. a TitleColumn)
    that has an element with the id "{obj_type}_{instance.pk}_title" that contains
    the title of the object (and nothing else) for screen reader purposes.
    """

    header_template_name = "wagtailadmin/bulk_actions/select_all_checkbox_cell.html"
    cell_template_name = "wagtailadmin/bulk_actions/listing_checkbox_cell.html"

    def __init__(self, *args, obj_type, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj_type = obj_type

    def get_aria_describedby(self, instance):
        return f"{self.obj_type}_{quote(instance.pk)}_title"

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context.update(
            {
                "obj_type": self.obj_type,
                "aria_describedby": self.get_aria_describedby(instance),
            }
        )
        return context


class UsageCountColumn(Column):
    cell_template_name = "wagtailadmin/tables/usage_count_column_cell.html"


class ReferencesColumn(Column):
    cell_template_name = "wagtailadmin/tables/references_cell.html"

    def __init__(
        self,
        name,
        label=None,
        accessor=None,
        classname=None,
        sort_key=None,
        width=None,
        get_url=None,
        describe_on_delete=False,
    ):
        super().__init__(name, label, accessor, classname, sort_key, width)
        self._get_url_func = get_url
        self.describe_on_delete = describe_on_delete

    def get_edit_url(self, instance):
        if self._get_url_func:
            return self._get_url_func(instance)

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["edit_url"] = self.get_edit_url(instance)
        context["describe_on_delete"] = self.describe_on_delete
        return context


class DownloadColumn(Column):
    cell_template_name = "wagtailadmin/tables/download_cell.html"

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["download_url"] = instance.url
        return context


class RelatedObjectsColumn(Column):
    """Outputs a list of objects related to the object through a one-to-many relationship"""

    cell_template_name = "wagtailadmin/tables/related_objects_cell.html"

    def get_value(self, instance):
        return getattr(instance, self.accessor).all()


class Table(Component):
    template_name = "wagtailadmin/tables/table.html"
    classname = "listing"
    header_row_classname = ""
    ascending_title_text_format = gettext_lazy(
        "Sort by '%(label)s' in ascending order."
    )
    descending_title_text_format = gettext_lazy(
        "Sort by '%(label)s' in descending order."
    )

    def __init__(
        self,
        columns,
        data,
        template_name=None,
        base_url=None,
        ordering=None,
        classname=None,
        attrs=None,
        caption=None,
    ):
        self.columns = OrderedDict([(column.name, column) for column in columns])
        self.caption = caption
        self.data = data
        if template_name:
            self.template_name = template_name
        self.base_url = base_url
        self.ordering = ordering
        if classname is not None:
            self.classname = classname
        self.base_attrs = attrs or {}

    def get_caption(self):
        return self.caption

    def get_context_data(self, parent_context):
        context = super().get_context_data(parent_context)
        context["table"] = self
        context["request"] = parent_context.get("request")
        return context

    @property
    def media(self):
        media = super().media
        for col in self.columns.values():
            media += col.media
        return media

    @property
    def rows(self):
        for index, instance in enumerate(self.data):
            yield Table.Row(self, instance, index)

    @property
    def row_count(self):
        return len(self.data)

    @property
    def attrs(self):
        attrs = self.base_attrs.copy()
        attrs["class"] = self.classname
        return attrs

    def get_row_classname(self, instance):
        return ""

    def get_row_attrs(self, instance):
        attrs = {}
        classname = self.get_row_classname(instance)
        if classname:
            attrs["class"] = classname
        return attrs

    def has_column_widths(self):
        return any(column.width for column in self.columns.values())

    def get_ascending_title_text(self, column):
        if self.ascending_title_text_format:
            return self.ascending_title_text_format % {"label": column.label}

    def get_descending_title_text(self, column):
        if self.descending_title_text_format:
            return self.descending_title_text_format % {"label": column.label}

    class Row(Mapping):
        # behaves as an OrderedDict whose items are the rendered results of
        # the corresponding column's format_cell method applied to the instance
        def __init__(self, table, instance, index):
            self.table = table
            self.columns = table.columns
            self.instance = instance
            self.index = index

        def __len__(self):
            return len(self.columns)

        def __getitem__(self, key):
            return self.columns[key].get_cell(self.instance)

        def __iter__(self):
            yield from self.columns

        def __repr__(self):
            return repr([col.get_cell(self.instance) for col in self.columns.values()])

        @cached_property
        def classname(self):
            return self.table.get_row_classname(self.instance)

        @cached_property
        def attrs(self):
            return self.table.get_row_attrs(self.instance)



import warnings

from wagtail.admin.telepath import (  # noqa: F401
    Adapter,
    AdapterRegistry,
    JSContext,
    JSContextBase,
    WagtailAdapterRegistry,
    WagtailJSContextBase,
    adapter,
    register,
    registry,
)
from wagtail.utils.deprecation import RemovedInWagtail80Warning

warnings.warn(
    "wagtail.telepath has been moved to wagtail.admin.telepath",
    RemovedInWagtail80Warning,
    stacklevel=2,
)



from django.db import models


class Orderable(models.Model):
    sort_order = models.IntegerField(null=True, blank=True, editable=False)
    sort_order_field = "sort_order"

    class Meta:
        abstract = True
        ordering = ["sort_order"]



from django.utils.html import format_html

from wagtail.embeds.exceptions import EmbedException, EmbedNotFoundException

from .base import EmbedFinder


class EmbedlyException(EmbedException):
    pass


class AccessDeniedEmbedlyException(EmbedlyException):
    pass


class EmbedlyFinder(EmbedFinder):
    key = None

    def __init__(self, key=None):
        if key:
            self.key = key

    def get_key(self):
        return self.key

    def accept(self, url):
        # We don't really know what embedly supports so accept everything
        return True

    def find_embed(self, url, max_width=None, key=None):
        from embedly import Embedly

        # Get embedly key
        if key is None:
            key = self.get_key()

        # Get embedly client
        client = Embedly(key=key)

        # Call embedly
        if max_width is not None:
            oembed = client.oembed(url, maxwidth=max_width, better=False)
        else:
            oembed = client.oembed(url, better=False)

        # Check for error
        if oembed.get("error"):
            if oembed["error_code"] in [401, 403]:
                raise AccessDeniedEmbedlyException
            elif oembed["error_code"] == 404:
                raise EmbedNotFoundException
            else:
                raise EmbedlyException

        # Convert photos into HTML
        if oembed["type"] == "photo":
            html = format_html('<img src="{}" alt="">', oembed["url"])
        else:
            html = oembed.get("html")

        # Return embed as a dict
        return {
            "title": oembed["title"] if "title" in oembed else "",
            "author_name": oembed["author_name"] if "author_name" in oembed else "",
            "provider_name": oembed["provider_name"]
            if "provider_name" in oembed
            else "",
            "type": oembed["type"],
            "thumbnail_url": oembed.get("thumbnail_url"),
            "width": oembed.get("width"),
            "height": oembed.get("height"),
            "html": html,
        }


embed_finder_class = EmbedlyFinder



from django.contrib.admin.utils import quote
from django.test import TestCase
from django.urls import reverse

from wagtail.test.testapp.models import ModelWithStringTypePrimaryKey
from wagtail.test.utils import WagtailTestUtils


class TestGenericIndexView(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def get(self, params={}):
        return self.client.get(reverse("testapp_generic_index"), params)

    def test_non_integer_primary_key(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        response_object_count = response.context_data["object_list"].count()
        self.assertEqual(response_object_count, 4)
        self.assertContains(response, "first modelwithstringtypeprimarykey model")
        self.assertContains(response, "second modelwithstringtypeprimarykey model")
        soup = self.get_soup(response.content)
        h1 = soup.select_one("h1")
        self.assertIsNotNone(h1)
        self.assertEqual(h1.text.strip(), "Model with string type primary keys")


class TestGenericIndexViewWithoutModel(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def get(self, params={}):
        return self.client.get(reverse("testapp_generic_index_without_model"), params)

    def test_non_integer_primary_key(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        response_object_count = response.context_data["object_list"].count()
        self.assertEqual(response_object_count, 4)


class TestGenericCreateView(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def get(self, params={}):
        return self.client.get(reverse("testapp_generic_create"), params)

    def test_get_create_view(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        h2 = soup.select_one("main h2")
        self.assertIsNotNone(h2)
        self.assertEqual(h2.text.strip(), "Model with string type primary key")
        form = soup.select_one("main form")
        self.assertIsNotNone(form)
        id_input = form.select_one("input[name='custom_id']")
        self.assertIsNotNone(id_input)
        self.assertEqual(id_input.get("type"), "text")
        content_input = form.select_one("input[name='content']")
        self.assertIsNotNone(content_input)

    def test_post_create_view(self):
        post_data = {
            "custom_id": "string-pk-3",
            "content": "third modelwithstringtypeprimarykey model",
        }
        response = self.client.post(reverse("testapp_generic_create"), post_data)
        self.assertEqual(response.status_code, 302)  # Redirect to index view
        self.assertTrue(
            ModelWithStringTypePrimaryKey.objects.filter(pk="string-pk-3").exists()
        )


class TestGenericEditView(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def get(self, object_pk, params={}):
        return self.client.get(
            reverse("testapp_generic_edit", args=(object_pk,)), params
        )

    def test_non_integer_primary_key(self):
        response = self.get("string-pk-2")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "second modelwithstringtypeprimarykey model")

    def test_non_url_safe_primary_key(self):
        object_pk = 'string-pk-:#?;@&=+$,"[]<>%'
        response = self.get(quote(object_pk))
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, "non-url-safe pk modelwithstringtypeprimarykey model"
        )

    def test_unquote_sensitive_primary_key(self):
        object_pk = "web_407269_1"
        response = self.get(quote(object_pk))
        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response, "unquote-sensitive modelwithstringtypeprimarykey model"
        )

    def test_using_quote_in_edit_url(self):
        for object_pk in ('string-pk-:#?;@&=+$,"[]<>%', "web_407269_1"):
            with self.subTest(object_pk=object_pk):
                response = self.get(quote(object_pk))
                edit_url = response.context_data["action_url"]
                edit_url_pk = edit_url.split("/")[-2]
                self.assertEqual(edit_url_pk, quote(object_pk))

    def test_using_quote_in_delete_url(self):
        for object_pk in ('string-pk-:#?;@&=+$,"[]<>%', "web_407269_1"):
            with self.subTest(object_pk=object_pk):
                response = self.get(quote(object_pk))
                delete_url = response.context_data["delete_url"]
                delete_url_pk = delete_url.split("/")[-2]
                self.assertEqual(delete_url_pk, quote(object_pk))


class TestGenericDeleteView(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def get(self, object_pk, params={}):
        return self.client.get(
            reverse("testapp_generic_edit", args=(object_pk,)), params
        )

    def test_with_non_integer_primary_key(self):
        response = self.get("string-pk-2")
        self.assertEqual(response.status_code, 200)

    def test_with_non_url_safe_primary_key(self):
        object_pk = 'string-pk-:#?;@&=+$,"[]<>%'
        response = self.get(quote(object_pk))
        self.assertEqual(response.status_code, 200)

    def test_with_unquote_sensitive_primary_key(self):
        object_pk = "web_407269_1"
        response = self.get(quote(object_pk))
        self.assertEqual(response.status_code, 200)



from django.core.exceptions import ImproperlyConfigured
from django.utils.safestring import mark_safe

from wagtail.admin.forms.models import (
    WagtailAdminDraftStateFormMixin,
    WagtailAdminModelForm,
)
from wagtail.admin.telepath import register as register_telepath_adapter
from wagtail.admin.ui.components import Component
from wagtail.blocks import StreamValue
from wagtail.coreutils import safe_snake_case
from wagtail.models import DraftStateMixin
from wagtail.rich_text import RichText
from wagtail.utils.text import text_from_html


def get_form_for_model(
    model,
    form_class=WagtailAdminModelForm,
    **kwargs,
):
    """
    Construct a ModelForm subclass using the given model and base form class. Any additional
    keyword arguments are used to populate the form's Meta class.
    """

    # This is really just Django's modelform_factory, tweaked to accept arbitrary kwargs.

    meta_class_attrs = kwargs
    meta_class_attrs["model"] = model

    # The kwargs passed here are expected to come from Panel.get_form_options, which collects
    # them by descending the tree of child edit handlers. If there are no edit handlers that
    # specify form fields, this can legitimately result in both 'fields' and 'exclude' being
    # absent, which ModelForm doesn't normally allow. In this case, explicitly set fields to [].
    if "fields" not in meta_class_attrs and "exclude" not in meta_class_attrs:
        meta_class_attrs["fields"] = []

    # Give this new form class a reasonable name.
    class_name = model.__name__ + "Form"
    bases = (form_class.Meta,) if hasattr(form_class, "Meta") else ()
    Meta = type("Meta", bases, meta_class_attrs)
    form_class_attrs = {"Meta": Meta}

    metaclass = type(form_class)
    bases = [form_class]
    if issubclass(model, DraftStateMixin):
        bases.insert(0, WagtailAdminDraftStateFormMixin)
    return metaclass(class_name, tuple(bases), form_class_attrs)


class Panel:
    """
    Defines part (or all) of the edit form interface for pages and other models
    within the Wagtail admin. Each model has an associated top-level panel definition
    (also known as an edit handler), consisting of a nested structure of ``Panel`` objects.
    This provides methods for obtaining a :class:`~django.forms.ModelForm` subclass,
    with the field list and other parameters collated from all panels in the structure.
    It then handles rendering that form as HTML.

    The following parameters can be used to customize how the panel is displayed.
    For more details, see :ref:`customizing_panels`.

    :param heading: The heading text to display for the panel.
    :param classname: A CSS class name to add to the panel's HTML element.
    :param help_text: Help text to display within the panel.
    :param base_form_class: The base form class to use for the panel. Defaults to the model's ``base_form_class``, before falling back to :class:`~wagtail.admin.forms.WagtailAdminModelForm`. This is only relevant for the top-level panel.
    :param icon: The name of the icon to display next to the panel heading.
    :param attrs: A dictionary of HTML attributes to add to the panel's HTML element.
    """

    BASE_ATTRS = {}

    def __init__(
        self,
        heading="",
        classname="",
        help_text="",
        base_form_class=None,
        icon="",
        attrs=None,
    ):
        self.heading = heading
        self.classname = classname
        self.help_text = help_text
        self.base_form_class = base_form_class
        self.icon = icon
        self.model = None
        self.attrs = self.BASE_ATTRS.copy()

        if attrs is not None:
            self.attrs.update(attrs)

    def clone(self):
        """
        Create a clone of this panel definition. By default, constructs a new instance, passing the
        keyword arguments returned by ``clone_kwargs``.
        """
        return self.__class__(**self.clone_kwargs())

    def clone_kwargs(self):
        """
        Return a dictionary of keyword arguments that can be used to create a clone of this panel definition.
        """
        return {
            "icon": self.icon,
            "attrs": self.attrs,
            "heading": self.heading,
            "classname": self.classname,
            "help_text": self.help_text,
            "base_form_class": self.base_form_class,
        }

    def get_form_options(self):
        """
        Return a dictionary of attributes such as 'fields', 'formsets' and 'widgets'
        which should be incorporated into the form class definition to generate a form
        that this panel can use.
        This will only be called after binding to a model (i.e. self.model is available).
        """
        return {}

    def get_form_class(self):
        """
        Construct a form class that has all the fields and formsets named in
        the children of this edit handler.
        """
        form_options = self.get_form_options()
        # If a custom form class was passed to the panel, use it.
        # Otherwise, use the base_form_class from the model.
        # If that is not defined, use WagtailAdminModelForm.
        model_form_class = getattr(self.model, "base_form_class", WagtailAdminModelForm)
        base_form_class = self.base_form_class or model_form_class

        return get_form_for_model(
            self.model,
            form_class=base_form_class,
            **form_options,
        )

    def bind_to_model(self, model):
        """
        Create a clone of this panel definition with a ``model`` attribute pointing to the linked model class.
        """
        new = self.clone()
        new.model = model
        new.on_model_bound()
        return new

    def get_bound_panel(self, instance=None, request=None, form=None, prefix="panel"):
        """
        Return a ``BoundPanel`` instance that can be rendered onto the template as a component. By default, this creates an instance
        of the panel class's inner ``BoundPanel`` class, which must inherit from ``Panel.BoundPanel``.
        """
        if self.model is None:
            raise ImproperlyConfigured(
                "%s.bind_to_model(model) must be called before get_bound_panel"
                % type(self).__name__
            )

        if not issubclass(self.BoundPanel, Panel.BoundPanel):
            raise ImproperlyConfigured(
                "%s.BoundPanel must be a subclass of Panel.BoundPanel"
                % type(self).__name__
            )

        return self.BoundPanel(
            panel=self, instance=instance, request=request, form=form, prefix=prefix
        )

    def on_model_bound(self):
        """
        Called after the panel has been associated with a model class and the ``self.model`` attribute is available;
        panels can override this method to perform additional initialisation related to the model.
        """
        pass

    def __repr__(self):
        return "<{} with model={}>".format(
            self.__class__.__name__,
            self.model,
        )

    def classes(self):
        """
        Additional CSS classnames to add to whatever kind of object this is at output.
        Subclasses of Panel should override this, invoking super().classes() to
        append more classes specific to the situation.
        """
        if self.classname:
            return [self.classname]
        return []

    def id_for_label(self):
        """
        The ID to be used as the 'for' attribute of any <label> elements that refer
        to this object but are rendered outside of it. Leave blank if this object does not render
        as a single input field.
        """
        return ""

    @property
    def clean_name(self):
        """
        A name for this panel, consisting only of ASCII alphanumerics and underscores, suitable for use in identifiers.
        Usually generated from the panel heading. Note that this is not guaranteed to be unique or non-empty; anything
        making use of this and requiring uniqueness should validate and modify the return value as needed.
        """
        return safe_snake_case(self.heading)

    def format_value_for_display(self, value):
        """
        Hook to allow formatting of raw field values (and other attribute values) for human-readable
        display. For example, if rendering a ``RichTextField`` value, you might extract text from the HTML
        to generate a safer display value.
        """
        # Improve representation of many-to-many values
        if callable(getattr(value, "all", "")):
            return ", ".join(str(obj) for obj in value.all()) or "None"

        # Avoid rendering potentially unsafe HTML mid-form
        if isinstance(value, (RichText, StreamValue)):
            return text_from_html(value)

        return value

    @register_telepath_adapter
    class BoundPanel(Component):
        """
        A template component for a panel that has been associated with a model instance, form, and request.
        """

        def __init__(self, panel, instance, request, form, prefix):
            #: The panel definition corresponding to this bound panel
            self.panel = panel

            #: The model instance associated with this panel
            self.instance = instance

            #: The request object associated with this panel
            self.request = request

            #: The form object associated with this panel
            self.form = form

            #: A unique prefix for this panel, for use in HTML IDs
            self.prefix = prefix

            self.heading = self.panel.heading
            self.help_text = self.panel.help_text

        @property
        def classname(self):
            return self.panel.classname

        def classes(self):
            return self.panel.classes()

        @property
        def attrs(self):
            return self.panel.attrs

        @property
        def icon(self):
            return self.panel.icon

        def id_for_label(self):
            """
            Returns an HTML ID to be used as the target for any label referencing this panel.
            """
            return self.panel.id_for_label()

        def is_shown(self):
            """
            Whether this panel should be rendered; if false, it is skipped in the template output.
            """
            return True

        def show_panel_furniture(self):
            """
            Whether this panel shows the panel furniture instead of being rendered outside of it.
            """
            return self.is_shown()

        def is_required(self):
            return False

        def get_context_data(self, parent_context=None):
            context = super().get_context_data(parent_context)
            context["self"] = self
            context["attrs"] = self.attrs
            return context

        def get_comparison(self):
            return []

        def render_missing_fields(self):
            """
            Helper function: render all of the fields that are defined on the form but not "claimed" by
            any panels via required_fields. These fields are most likely to be hidden fields introduced
            by the forms framework itself, such as ORDER / DELETE fields on formset members.
            (If they aren't actually hidden fields, then they will appear as ugly unstyled / label-less fields
            outside of the panel furniture. But there's not much we can do about that.)
            """
            rendered_fields = self.panel.get_form_options().get("fields", [])
            missing_fields_html = [
                str(self.form[field_name])
                for field_name in self.form.fields
                if field_name not in rendered_fields
            ]

            return mark_safe("".join(missing_fields_html))

        def render_form_content(self):
            """
            Render this as an 'object', ensuring that all fields necessary for a valid form
            submission are included
            """
            return mark_safe(self.render_html() + self.render_missing_fields())

        def __repr__(self):
            return "<{} with model={} instance={} request={} form={}>".format(
                self.__class__.__name__,
                self.panel.model,
                self.instance,
                self.request,
                self.form.__class__.__name__,
            )

        telepath_adapter_name = "wagtail.panels.Panel"

        def js_opts(self):
            return {
                "type": type(self.panel).__name__,
                "prefix": self.prefix,
            }

        def telepath_pack(self, context):
            return (self.telepath_adapter_name, [self.js_opts()])



# Generated by Django 1.10.4 on 2016-12-16 17:15
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0017_reduce_focal_point_key_max_length"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="rendition",
            name="filter",
        ),
    ]



import time
import tracemalloc


class Benchmark:
    repeat = 10

    def test(self):
        timings = []
        memory_usage = []
        tracemalloc.start()

        for i in range(self.repeat):
            before_memory = tracemalloc.take_snapshot()
            start_time = time.time()

            self.bench()

            end_time = time.time()
            after_memory = tracemalloc.take_snapshot()
            timings.append(end_time - start_time)
            memory_usage.append(
                sum(
                    [t.size for t in after_memory.compare_to(before_memory, "filename")]
                )
            )

        print(  # noqa: T201
            "time min:",
            min(timings),
            "max:",
            max(timings),
            "avg:",
            sum(timings) / len(timings),
        )
        print(  # noqa: T201
            "memory min:",
            min(memory_usage),
            "max:",
            max(memory_usage),
            "avg:",
            sum(memory_usage) / len(memory_usage),
        )
